from tkinter import *
from tkinter.ttk import *
from tkinter.tix import *

import matplotlib
matplotlib.use("TkAgg")
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2TkAgg
from matplotlib.figure import Figure
import matplotlib.animation as animation
import datetime

import time
import random
from collections import deque
from itertools import permutations, islice, product, combinations_with_replacement
import numpy as np
import statistics as stat
from functools import partial
import os.path

# ><^

# Globals
codes_file_name = "codes.dat"       # Name of the file where codes are saved
tables_file_name = "tables.dat"     # Name of the file where routing tables are saved
batch_file_name = "batches.dat"     # Name of the file where batch tables are saved
existing_codes = {}                 # Dictionary to store codes. Each code is a dict of {'name', 'width', 'height', 'matrix'}
existing_routing_tables = {}        # Dictionary to store routing tables. Each table is a dict of {'name', 'table'}
existing_batch_tables = {}          # Dictionary to store batch tables. {'name', 'batch_size', 'batch_table'}
selected_code = {}                  # A helper variable, used to display code info, etc. Contains the info of the code that is selected from the dropdown
recursion_helper = {}               # Used to store valid server combinations, which will make up the routing table
running_codes = []                  # A list of codes that are currently running
selected_statistic = ""             # A helper variable. Holds the value of the selected dropdown item
dir_path = os.path.dirname(os.path.realpath(__file__))

# List of collected statistics, used to draw graphs, define dropdown selections, etc.
statistics = [
    "request_queue_size",
    "current_requests",
    "completed_requests",
    "requests_per_second",
    "servers_in_use",
    "servers_at_capacity",
    "servers_capacity_percent",
    "queue_time_mean",
    "queue_time_median",
    "queue_time_stdev",
    "queue_time_variance",
    "queue_time_max"]

# Graph for displaying statistics
fig = Figure(figsize=(5,5), dpi=100)
graph = fig.add_subplot(111)

class Code:
    def __init__(self, name, width, height, matrix, server_values):
        global statistics, existing_rouuting_tables, existing_batch_tables
        self.paused = True                      # Determines if the code is actually running or not
        
        # Code defining parameters
        self.name = name                        # Name of the code
        self.server_amount = width              # Amount of servers for the code, defined by the width of the matrix
        self.file_amount = height               # Amount of files for the code, defined by the height of the matrix
        self.matrix = matrix                    # Matrix for the code
        self.server_values = server_values      # Values for the servers of the code (list of columns of the matrix)
        self.routing_table = {}                 # "Routing table" for the code. Shows what combinations of servers can be used to serve requests
        self.batch_table = {}                   # Batch table for the code
        self.simplex = True                     # Is the code a simplex code
        self.TTL_exponential = True             # Sets TTL distribution together with TTL_poisson & TTL_uniform
        self.TTL_poisson = False                # (Only one of the TTL distributions can be True at once)
        self.TTL_uniform = False
        self.TTL_gaussian = False
        self.run_as_batch = False               # Determines whether the code is run as a batch or asynchronous batch code
        
        # Code running configurations
        self.min_TTL = 1                        # Defines the minimum "time to live" for download requests
        self.max_TTL = 5                        # Defines the maximum "time to live" for download requests
        self.TTL_lambda = 3                     # Defines the Poisson probability distribution paramater lambda used to determine the TTL of new requests
        self.TTL_scale = 1                      # Defines the exponential distribution variable Beta = 1/lambda (as defined in numpy.random.exponential)
        self.TTL_mu = 1
        self.TTL_sigma = 0.1                    
        self.server_capacity = 1                # Defines the maximum amount of requests servers can handle
        self.batch_size = 1                     # Defines the batch size being used when running this code as a batch code
        self.skip_distance = 8                  # Defines how many requests are looked at for servability when running in async mode
        self.request_gap_lambdas = {}           # Holds the individual Poisson lambdas defining the gaps between requests for each file
        self.request_gap_time_left = {}         # Holds the times left until new request
        
        
        # Data variables for runnning the code
        self.request_queue = deque([])          # Queue for storing unserved requests
        self.current_requests = []              # List of requests currently being served
        self.completed_requests = 0             # Amount of completed requests
        self.servers_at_max_capacity = 0        # Amount of servers at maximum capacity (as defined by server_capacity)
        self.servers = {}                       # Dictionary for storing server information
        self.running_time = 0                   # Counter for "steps" (1 sec each) for which the code has been running
        self.statistics = {}                    # Dictionary of collected statistics. Used to store data for drawing graphs
        self.max_capacity = self.server_capacity * self.server_amount
        self.last_batch_failed = False

        # Set gap values at 0.2s
        self.request_gap_lambdas = {x: 0.2 for x in range(self.file_amount)}
        
        # Initiate statistics dictonary, use the global statistics list to get the names
        for stat in statistics:
            self.statistics[stat] = deque([], 6000)
            
        # Initiate the servers dictionary
        for server in range(self.server_amount):
            self.servers[server] = {"requests": 0}

        # If routing table exists for the code, read it
        if self.name in existing_routing_tables.keys():
            self.routing_table = existing_routing_tables[self.name]["table"]

        # If batch table exists for the code, read it
        if self.name in existing_batch_tables.keys():
            self.batch_table = existing_batch_tables[self.name]["batch_tables"]
            self.batch_size = max(self.batch_table.keys())

        # Check if code is simplex
        simplex_cols = list(product([0,1], repeat=self.file_amount))[1:]
        server_cols = []
        for col in self.server_values.keys():
            server_cols.append(self.server_values[col])
        for col in simplex_cols:
            if list(col) not in server_cols:
                self.simplex = False
                print("Not simplex")
                break

    def update_server_max_capacity(self):
        self.max_capacity = self.server_capacity * self.server_amount
        
    def set_TTL_exponential(self):
        self.TTL_poisson = False
        self.TTL_uniform = False
        self.TTL_exponential = True
        self.TTL_gaussian = False
        print("Set code TTL as exponential")

    def set_TTL_poisson(self):
        self.TTL_poisson = True
        self.TTL_uniform = False
        self.TTL_exponential = False
        self.TTL_gaussian = False
        print("Set code TTL as poisson")

    def set_TTL_uniform(self):
        self.TTL_poisson = False
        self.TTL_uniform = True
        self.TTL_exponential = False
        self.TTL_gaussian = False
        print("Set code TTL as uniform")

    def set_TTL_gaussian(self):
        self.TTL_poisson = False
        self.TTL_uniform = False
        self.TTL_exponential = False
        self.TTL_gaussian = True
        print("Set code TTL as gaussian")

    def get_file_name(self):
        f_name = self.name
        if self.run_as_batch:
            f_name += " batch, size="
            f_name += str(self.batch_size)
        else:
            f_name += " async, skip="
            f_name += str(self.skip_distance)

        f_name += " cap=" + str(self.server_capacity)
        if self.TTL_uniform:
            f_name += " TTL=" + str(int(self.min_TTL)) + "-" + str(int(self.max_TTL))
        elif self.TTL_poisson:
            f_name += " TTL_lam=" + str(int(self.TTL_lambda))
        elif self.TTL_exponential:
            f_name += " TTL_scale=" + str(int(self.TTL_scale))
        elif self.TTL_gaussian:
            f_name += " TTL_mu=" + str(self.TTL_mu) + " TTL_sigma=" + str(self.TTL_sigma)
        
        lambdas = ""
        for file in list(self.request_gap_lambdas.keys()):
            lambdas += str(self.request_gap_lambdas[file]) + "+"
        lambdas = lambdas[:-1]
        f_name += " req_lambdas=" + lambdas
        f_name += ".dat"

        return f_name

    # Write stats to file
    def write_stats_to_file(self):
        # Create file name based on code params
        f_name = self.get_file_name()

        stat_line = ""
        for stat in self.statistics.keys():
            stat_line += str(self.statistics[stat][-1]) + ","
        stat_line = stat_line[:-1]

        # If new file, write column names. 
        if os.path.isfile(f_name):
            f = open(f_name, "a+")
            f.write(stat_line + "\n")
            f.close()
        else:
            cols = ""
            for stat in self.statistics.keys():
                cols += stat + ","
            cols = cols[:-1]
            f = open(f_name, "w+")
            f.write(cols + "\n")
            f.write(stat_line + "\n")
            f.close()
            

    # Used to check if a request is servable. Returns first viable option from the routing table if servable, False otherwise
    def request_servability(self, request):
        server_options = self.routing_table[request["requested_file"]]

        for option in server_options:
            servable = True
            for server in option:
                if self.servers[server]["requests"] == self.server_capacity:
                    servable = False

            if servable:
                return option

        return False

    # Generates the routing table for the code
    def generate_routing_table(self):
        global tables_file_name, existing_routing_tables
        
        # Find all possible server combinations for each file, save the combinations in the routing_tables.dat file & set current_routing_table
        combinations = self.get_server_combinations(self.server_values, self.file_amount, self.server_amount)

        dict_item = {"name": self.name, "table": combinations}
        
        tables_file = open(tables_file_name, "a")
        tables_file.write(str(dict_item) + "\n")
        tables_file.close()

        existing_routing_tables[self.name] = dict_item

        self.routing_table = combinations
        
        return

    # Calls a recursive function to find all possible combinations of servers which could be used to retrieve any given file
    def get_server_combinations(self, server_values, file_amount, server_amount):
        global recursion_helper
        servers = [i for i in range(server_amount)]
        files = [i for i in range(file_amount)]
        combinations = {}
        recursion_helper = {}

        # Convert displayable info to binary for calculations
        bin_server_values = {}
        for server in servers:
            bin_server_values[server] = self.bin_list_to_int(server_values[server])

        bin_file_values = {}
        for file in files:
            bin_file_list = [0]*file_amount
            bin_file_list[file] = 1
            bin_file_values[file] = self.bin_list_to_int(bin_file_list)

        # Find the combinations with fewer servers first by limiting the recursion depth
        for depth in range(1, file_amount):
            t = time.time()
            for file in files:
                new_combinations = self.recursive_search(servers, bin_server_values, bin_file_values, file, depth, set([]), 0)
                if len(new_combinations) != 0:
                    if file in combinations.keys():
                        combinations[file] += new_combinations
                    else:
                        combinations[file] = new_combinations
                else:
                    print("Total time for depth " + str(depth) + ": " + str(time.time() - t))
                    print("No new combinations for depth " + str(depth) + ", breaking..")
                    return combinations
            print("Total time for depth " + str(depth) + ": " + str(time.time() - t))
        return combinations
    
    # Recursively goes over all combinations of servers, returns all combinations which can be used to retrieve the given file
    def recursive_search(self, servers, server_values, file_values, file, subset_size, used_servers, current_state):
        global recursion_helper

        # Uses the global recursion_helper to check if a subset of the current state is a known combination. If yes, break recursion
        if file in recursion_helper.keys():
            for existing_combination in recursion_helper[file]:
                if existing_combination.issubset(used_servers):
                    return []

        # Current state produces a valid combination of servers for the given file. Return the state and store it in global recursion_helper
        if current_state == file_values[file]:
            if file in recursion_helper.keys():
                recursion_helper[file].append(used_servers)
            else:
                recursion_helper[file] = [used_servers]
            return [used_servers]

        # If maximum recursion depth has been reached, break recursion
        if subset_size == len(used_servers):
            return []
        
        # Run the recursion again for all remaining servers
        combo = []
        for server in servers:
            new_state = self.XOR(current_state, server_values[server])
            new_servers = servers.copy()
            new_servers.remove(server)
            used_servers_copy = used_servers.copy()
            used_servers_copy.add(server)
            combo += self.recursive_search(new_servers, server_values, file_values, file, subset_size, used_servers_copy, new_state)

        return combo

    # Used to turn a list of 1s and 0s into actual binary (int)
    def bin_list_to_int(self, bitlist):
        out = 0
        for bit in bitlist:
            out = (out << 1) | bit
        return out

    def XOR(self, a, b):
        return a ^ b
        
    # Generates the batch table for the code
    def generate_batch_table(self):
        global batch_file_name, existing_batch_tables
        file_list = [i for i in range(self.file_amount)]
        batch_dicts = {}

        if self.simplex:
            batch_size = 2**(self.file_amount - 1)
        else:
            batch_size = 1 
            while True:
                permutation = [0]*batch_size
                used_servers = []
                print("Trying batch size " + str(batch_size))
                combinations = []
                for file in permutation:
                    combination = self.get_server_combination(used_servers, file)
                    
                    if not combination:
                        print("No combination found for batch_size " + str(len(permutation)))

                    if combination:
                        combinations.append(combination)
                        for server in combination:
                            used_servers.append(server)

                if not combination:
                    break
                print(combinations)
                batch_size += 1

            batch_size += -1
        print("Maximum batch size found: " + str(batch_size))
        self.batch_size = batch_size

        if batch_size <= 16:
            #batch_dict = self.recursive_batch_search(self.batch_size, [])
            batch_dict = self.non_recursive_batch_search(self.batch_size)
            batch_dicts[batch_size] = batch_dict
        else:
            batch_dicts[batch_size] = False
            
        dict_item = {"name": self.name, "batch_tables": batch_dicts}

        batch_file = open(batch_file_name, "a")
        batch_file.write(str(dict_item) + "\n")
        batch_file.close()

        existing_batch_tables[self.name] = dict_item
        self.batch_table = dict_item["batch_tables"]
            
        return dict_item
    
    # Recursively generates permutations with repetition for batches
    def recursive_batch_search(self, batch_size, permutation):
        # If batch_size reached, find suitable combination & return it
        if len(permutation) == batch_size:
            dict_key = ""
            used_servers = []
            combinations = []
            
            for file in permutation:
                dict_key += str(file)
                combination = self.get_server_combination(used_servers, file)
                
                if not combination:
                    print("No combination found for batch_size " + str(len(permutation)))
                    print(permutation)
                    return False
                
                for server in combination:
                    used_servers.append(server)
                combinations.append(combination)

            batch_dict = {dict_key: combinations}
            return batch_dict
        # If batch_size not reached, continue recursion until done
        else:
            batch_dict = {}
            for file in [i for i in range(self.file_amount)]:
                res = self.recursive_batch_search(batch_size, permutation + [file])
                if res == False:
                    return False
                else:
                    batch_dict.update(res)

        return batch_dict

    # Assigns servers to all unique sets of requests (with size = batch_size)
    def non_recursive_batch_search(self, batch_size):
        unique_sets = list(combinations_with_replacement(list(range(self.file_amount)), batch_size))
        batch_table = {}
        
        print("Unique sets: " + str(len(unique_sets)))
        for unique_perm in unique_sets:
            name = ""
            for file in unique_perm:
                name += str(file)
            
            perm = list(unique_perm)
            comb = False
            while not comb:
                used_servers = []
                combinations = []
                
                for file in perm:
                    combination = self.get_server_combination(used_servers, file)
                    if combination:
                        for server in combination:
                            used_servers.append(server)
                        combinations.append(combination)
                    else:
                        random.shuffle(perm)

                if combinations:
                    if len(combinations) == batch_size:
                        comb = combinations

            ordered_comb = []
            for i in unique_perm:
                for j in perm:
                    if i == j:
                        ordered_comb.append(combinations[perm.index(j)])
                        combinations.pop(perm.index(j))
                        perm.remove(j)
                        break
            
            batch_table[name] = ordered_comb
        return batch_table
        
    # Helper function for generating server combinations for batch tables
    def get_server_combination(self, used_servers, file):
        for combination in self.routing_table[file]:
            valid_combination = True
            for server in combination:
                if server in used_servers:
                    valid_combination = False
            if valid_combination:
                return combination
        return False
    
    # Checks if a batch of requests is servable
    def batch_servable(self, servers):
        servable = True
        for server in servers:
            if self.servers[server]["requests"] == self.server_capacity:
                servable = False
        if servable:
            return True
        return False

    # Serves a batch
    def serve_batch(self, batch, servers, batch_combinations):
        counter = 0
        #print(batch_combinations)
        for request in batch:
            request["servers"] = batch_combinations[counter]
            self.current_requests.append(request)
            counter += 1
        for server in servers:
            self.servers[server]["requests"] += 1

    # Generates a new request for the given file and does all the surrounding procedures
    def generate_new_request(self, file):
        request = {}
        request["requested_file"] = file
        request["servers"] = []
        request["time_in_queue"] = 0
        
        # Determine TTL
        if self.TTL_uniform:
            request["TTL"] = np.random.random_integers(self.min_TTL*100, self.max_TTL*100)
        elif self.TTL_poisson:
            request["TTL"] = np.random.poisson(self.TTL_lambda*100)
        elif self.TTL_exponential:
            request["TTL"] = (np.random.exponential(self.TTL_scale*100))
        elif self.TTL_gaussian:
            request["TTL"] = np.random.normal(self.TTL_mu*100, self.TTL_sigma*100)

        # Get the time until next request & set the gap
        gap = np.random.exponential(self.request_gap_lambdas[file]*100)
        self.request_gap_time_left[file] = gap - 1

        #print("New request for F" + str(file + 1) + " at t=" + str(self.running_time/100) + "sec. TTL=" + str(round(request["TTL"]/100, 1)) + "sec, gap=" + str(gap/100) + "sec")
    
        self.request_queue.append(request)

    # Sorts the requests in the batch according to batch_name to make sure that correct servers are assigned to requests
    def order_batch(self, batch, batch_name):
        batch_copy = batch.copy()
        ordered_batch = []
        for char in batch_name:
            for request in batch_copy:
                if int(char) == request["requested_file"]:
                    ordered_batch.append(request)
                    batch_copy.remove(request)
                    break
        return ordered_batch

    # Runs a "step" for the code (equivalent to 1 second passing while the code is running)
    def run_code_step(self):
        self.running_time += 1
        
        # Check the "time to live" (TTL) of each request being served to see if they are done
        living_requests = []
        for request in self.current_requests:
            request["TTL"] = request["TTL"] - 1
            if request["TTL"] < 1:
                self.completed_requests += 1
                for server in request["servers"]:
                    self.servers[server]["requests"] += -1
            else:
                living_requests.append(request)

        self.current_requests = living_requests

        # Try to serve new requests
        # If code is asynchronous
        if not self.run_as_batch:
            first_servable = True
            while first_servable:
                if len(self.request_queue) != 0:
                    request = self.request_queue[0]
                    serve = self.request_servability(request)
                    if not serve:
                        first_servable = False
                    else:
                        self.request_queue.remove(request)
                        request["servers"] = serve
                        for server in request["servers"]:
                            self.servers[server]["requests"] += 1
                        self.current_requests.append(request)
                else:
                    first_servable = False
            
            # Once top of the queue has been served (until unservable request), skip over skip_distance elements to see if other requests could be served
            batch = list(islice(self.request_queue, 0, self.skip_distance))
            for request in batch:
                serve = self.request_servability(request)
                if serve != False:
                    self.request_queue.remove(request)
                    request["servers"] = serve
                    for server in request["servers"]:
                        self.servers[server]["requests"] += 1
                    self.current_requests.append(request)
        # If code is regular batch code
        else:
            if len(self.request_queue) >= self.batch_size:
                if not (self.server_capacity == 1 and len(self.current_requests) != 0):
                    batch = list(islice(self.request_queue, 0, self.batch_size))
                    batch_name = ""

                    if self.last_batch_failed:
                        random.shuffle(batch)
                        for request in batch:
                            batch_name += str(request["requested_file"])
                        print("shuffled: " + str(batch_name))
                        self.last_batch_failed = False

                    batch_list = []
                    for request in batch:
                        batch_list.append(str(request["requested_file"]))
                    batch_list.sort()
                    batch_name = "".join(batch_list)
                    batch = self.order_batch(batch, batch_name)
                    
                    # If the code has a batch table
                    if self.batch_table[self.batch_size]:
                        batch_combinations = self.batch_table[self.batch_size][batch_name]
                        servers = []

                        for combination in batch_combinations:
                            for server in combination:
                                servers.append(server)

                        if self.batch_servable(servers):
                            #print("serving batch: " + batch_name)
                            self.serve_batch(batch, servers, batch_combinations)
                            for request in batch:
                                self.request_queue.remove(request)
                       
        # Add +1 to "time_in_queue" for queued requests. Add all times to list for doing statistics
        queue_times = []
        for request in self.request_queue:
            request["time_in_queue"] += 1
            queue_times.append(request["time_in_queue"]/100)

        # Use the Python Statistics library to analyze queue times
        try:
            queue_mean = stat.mean(queue_times)
        except:
            queue_mean = 0

        try:
            queue_median = stat.median(queue_times)
        except:
            queue_median = 0

        try:
            queue_stdev = stat.stdev(queue_times)
        except:
            queue_stdev = 0

        try:
            queue_variance = stat.variance(queue_times)
        except:
            queue_variance = 0

        try:
            queue_max = max(queue_times)
        except:
            queue_max = 0
                
        # Add server capacity stats
        servers_in_use = 0
        servers_at_capacity = 0
        used_capacity = 0
            
        for server in self.servers.keys():
            used_capacity += self.servers[server]["requests"]
            if self.servers[server]["requests"] != 0:
                servers_in_use += 1
            if self.servers[server]["requests"] == self.server_capacity:
                servers_at_capacity += 1

        servers_capacity_percent = used_capacity / (self.max_capacity / 100)
        requests_per_second = self.completed_requests / (self.running_time / 100)

        # Add stats to the global variable "running_codes_statistics"
        self.statistics["request_queue_size"].append(len(self.request_queue))
        self.statistics["current_requests"].append(len(self.current_requests))
        self.statistics["completed_requests"].append(self.completed_requests)
        self.statistics["requests_per_second"].append(requests_per_second)
        self.statistics["servers_in_use"].append(servers_in_use)
        self.statistics["servers_at_capacity"].append(servers_at_capacity)
        self.statistics["servers_capacity_percent"].append(servers_capacity_percent)
        self.statistics["queue_time_mean"].append(queue_mean)
        self.statistics["queue_time_median"].append(queue_median)
        self.statistics["queue_time_stdev"].append(queue_stdev)
        self.statistics["queue_time_variance"].append(queue_variance)
        self.statistics["queue_time_max"].append(queue_max)

        # Generate new requests if its time to
        for file in range(self.file_amount):
            if file in self.request_gap_time_left.keys():
                if self.request_gap_time_left[file] < 1:
                    self.generate_new_request(file)
                else:
                    self.request_gap_time_left[file] += -1
            else:
                self.generate_new_request(file)            
        
        self.write_stats_to_file()
        
# Animates the graph by redrawing it whenever called. Uses the globals "selected_statistic" & "running_codes" to get data for plotting
def animate(i):
    global selected_statistic
    graph.clear()
    # x-axis is time (1min)
    xList = range(1, 6001)
    for code in running_codes:
        yList = list(code.statistics[selected_statistic])
        if len(yList) != 6000:
            yList = [0]*(6000 - len(yList)) + yList
        graph.plot(xList, yList, label = code.name)
        graph.set_title(selected_statistic)
        graph.legend(loc = 2)

# Main method used to run the program GUI and update the graph at 1-second intervals
def main():
    get_existing_codes()
    root = Tk()
    gui = MainWindow(root)
    ani = animation.FuncAnimation(fig, animate, interval=1000)
    root.mainloop()
    
# Reads codes and tables from files, if they exist
def get_existing_codes():
    global existing_codes, existing_routing_tables, existing_batch_tables
    try:
        codes_file = open(codes_file_name, "r")
        for line in codes_file.readlines():
            dict_item = eval(line)
            name = dict_item["name"]
            existing_codes[name] = dict_item
        codes_file.close()
    except FileNotFoundError:
        return  
    try:
        tables_file = file = open(tables_file_name, "r")
        for line in tables_file.readlines():
            dict_item = eval(line)
            name = dict_item["name"]
            existing_routing_tables[name] = dict_item
        tables_file.close()
    except FileNotFoundError:
        return
    try:
        batch_file = open(batch_file_name, "r")
        for line in batch_file.readlines():
            dict_item = eval(line)
            name = dict_item["name"]
            existing_batch_tables[name] = dict_item
        batch_file.close()
    except FileNotFoundError:
        return

# Main GUI window, contains all the displayable info - the File\Server matrix, routing table and graph, etc.
class MainWindow:
    def __init__(self, master):
        global statistics, selected_code
        # Main window
        self.master = master
        self.master.after(1000, self.run_codes)
        master.title("MainWindow")
        self.label = Label(self.master, text="MainWindow")

        # Frames
        self.dropdown_frame = Frame(self.master)
        self.dropdown_frame.pack()
        self.running_codes_frame = Frame(self.master)
        self.running_codes_frame.pack()
        self.graph_frame = Frame(self.master)
        self.graph_frame.pack()
        
        # Codes Dropdown Label
        self.dropdown_label = Label(self.dropdown_frame, text="Select code: ", anchor=W, justify=LEFT)
        self.dropdown_label.grid(column=0, row=0)

        # Codes Dropdown
        self.dropdown_str = StringVar()
        self.dropdown_str.trace('w', self.dropdown_item_selected)
        self.dropdown = Combobox(self.dropdown_frame, textvariable=self.dropdown_str, width=60)
        values = [""]
        if len(list(existing_codes.keys())) != 0:
            values = list(existing_codes.keys())
        self.dropdown['values']= values
        self.dropdown.current(0)
        self.dropdown.grid(column=1, row=0)

        # GO Button
        self.go_btn = Button(self.dropdown_frame, text="GO", command=self.go_btn_clicked, width=8)
        self.go_btn.grid(column=2, row=0)

        # New Code Button
        self.new_code_btn = Button(self.dropdown_frame, text="Create new", command=self.new_code_btn_clicked, width=8)
        self.new_code_btn.grid(column=3, row=0)

        # GraphFrame
        self.stat_dropdown_str = StringVar()
        self.stat_dropdown_str.trace('w', self.stat_dropdown_item_selected)
        self.stat_dropdown = Combobox(self.graph_frame, textvariable=self.stat_dropdown_str, width=60)
        self.stat_dropdown['values']= statistics
        self.stat_dropdown.current(0)
        self.stat_dropdown.pack()
        self.graph_frame_content = GraphFrame(self.graph_frame, self)
        self.graph_frame_content.pack()

        # Set the global selected_code
        selected_code = existing_codes[self.dropdown_str.get()]
            
        
    # Used to display running codes when the global running_codes is changed
    def update_running_codes_frame(self):
        global running_codes
        # Clear the frame
        for entry in self.running_codes_frame.winfo_children():
            entry.destroy()

        row_counter = 0
        for code in running_codes:
            # Code name label
            code_name_str = code.name
            if code.run_as_batch:
                code_name_str += " batch, size: " + str(code.batch_size)
            else:
                code_name_str += " async, skip: " + str(code.skip_distance)
            code_name = Label(self.running_codes_frame, text=code_name_str, width=30)
            code_name.grid(row=row_counter, column=0)

            # Pause/Unpause button
            if code.paused:
                paused_txt = "Unpause"
            else:
                paused_txt = "Pause"
            pause_button = Button(self.running_codes_frame, text=paused_txt, command=partial(self.pause_button_clicked, code), width=6)
            pause_button.grid(row=row_counter, column=1)            

            # Settings button
            settings_button = Button(self.running_codes_frame, text="Settings", command=partial(self.settings_button_clicked, code), width=6)
            settings_button.grid(row=row_counter, column=2)

            # Remove button
            remove_button = Button(self.running_codes_frame, text="Remove", command=partial(self.remove_button_clicked, code), width=6)
            remove_button.grid(row=row_counter, column=3)

            # Info button
            info_button = Button(self.running_codes_frame, text="Info", command=partial(self.info_button_clicked, code), width=6)
            info_button.grid(row=row_counter, column=4)

            # Requests button
            requests_button = Button(self.running_codes_frame, text="Requests", command=partial(self.requests_button_clicked, code), width=6)
            requests_button.grid(row=row_counter, column=5)
            
            row_counter += 1

        self.running_codes_frame.pack()
    
        
    # Used to display code info, if a code is selected from the dropdown
    def dropdown_item_selected(self, index, value, op):
        global selected_code, existing_codes
        selected_code = existing_codes[self.dropdown_str.get()]

    # Triggered when statistics dropdown item selected.
    def stat_dropdown_item_selected(self, index, value, op):
        global selected_statistic
        stat = self.stat_dropdown.get()
        selected_statistic = stat

    # Sets the paused variable for the code and updates running_codes_frame
    def pause_button_clicked(self, code):
        if code.paused:
            code.paused = False
            print("Unpausing " + code.name)
        else:
            code.paused = True
            print("Pausing " + code.name)
        self.update_running_codes_frame()

    # Opens the settings window for the code
    def settings_button_clicked(self, code):
        self.code_settings_window = Toplevel(self.master)
        self.code_settings = SettingsWindow(self.code_settings_window, code)
        
    # Removes the code from running_codes
    def remove_button_clicked(self, code):
        global running_codes
        running_codes.remove(code)
        self.update_running_codes_frame()

    # Opens the info window for the code
    def info_button_clicked(self, code):
        self.code_info_window = Toplevel(self.master)
        self.code_info = InfoWindow(self.code_info_window, code)

    # Opens the info window for the code
    def requests_button_clicked(self, code):
        self.current_requests_window = Toplevel(self.master)
        self.current_requests = CurrentRequests(self.current_requests_window, code)
        
    # Starts main_loop for the selected code when GO button is pressed
    def go_btn_clicked(self):
        global running_codes, selected_code, existing_routing_tables
        name = selected_code["name"]
        width = selected_code["width"]
        height = selected_code["height"]
        matrix = selected_code["matrix"]
        server_values = selected_code["server_values"]
        print("Added code to running_codes: " + name)
        code = Code(name, width, height, matrix, server_values)
        running_codes.append(code)
        self.update_running_codes_frame()

    # Opens window to create new code when New Code button is pressed
    def new_code_btn_clicked(self):
        self.code_creator_window = Toplevel(self.master)
        self.code_creator = CodeCreator(self.code_creator_window)

        values = list(existing_codes.keys())
        self.dropdown['values']= values
        self.dropdown.current(values.index(selected_code["name"]))
        

    # Called by the window. Simulates a running code
    def run_codes(self):
        global running_codes, testing_3x7, testing_3x7_list, testing_times, testing_gaussian, gaussian_list, testing_skip, skip_list, testing_rep, rep_list, compare_models, constant_lifetime
        self.master.after(10, self.run_codes)
        
        # For each code in running_codes, run a code step if the code is not paused
        for code in running_codes:
            if not code.paused:
                code.run_code_step()
            
        # If no codes are running, just return
        if len(running_codes) == 0:
            return

    # Returns printable version of time from seconds
    def seconds_to_displayable(self, s):
        return str(datetime.timedelta(seconds=s))
        
                

# Matrix creation window                
class CodeCreator:
    def __init__(self, master):
        # Frames
        self.master = master
        self.name_frame = Frame(self.master)
        self.button_frame = Frame(self.master)
        self.entry_frame = Frame(self.master)
        self.matrix_frame = Frame(self.master)
        self.name_frame.pack()
        self.button_frame.pack()
        self.entry_frame.pack()

        # Buttons
        self.done_btn = Button(self.button_frame, text = 'Done', width = 8, command = self.done_btn_pressed)
        self.done_btn.grid(column=0, row=0)
        self.cancel_button = Button(self.button_frame, text = 'Cancel', width = 8, command = self.master.destroy)
        self.cancel_button.grid(column=1, row=0)
        
        # Name Label
        self.name_lbl = Label(self.name_frame, text="Name: ", anchor=W, justify=LEFT)
        self.name_lbl.grid(column=0, row=0)

        # Name Entry
        self.name_entry = Entry(self.name_frame, width=15)
        self.name_entry.grid(column=1, row=0)
        
        # Width Label
        self.width_lbl = Label(self.entry_frame, text="Width: ", anchor=W, justify=LEFT)
        self.width_lbl.grid(column=0, row=1)

        # Width Entry
        self.width_str = StringVar()
        self.width_str.trace("w", self.size_changed)
        self.width_entry = Entry(self.entry_frame, textvariable=self.width_str, width=15)
        self.width_entry.grid(column=1, row=1)

        # Height Label
        self.height_lbl = Label(self.entry_frame, text="Height: ", anchor=W, justify=LEFT)
        self.height_lbl.grid(column=0, row=2)

        # Height Entry
        self.height_str = StringVar()
        self.height_str.trace("w", self.size_changed)
        self.height_entry = Entry(self.entry_frame, textvariable=self.height_str, width=15)
        self.height_entry.grid(column=1, row=2)

        self.master.wait_window()



    # Gets the values from entries in the matrix frame
    def get_matrix(self, frame):
        ones = []
        for entry in frame.children.values():
            info = entry.grid_info()
            row = int(info["row"])
            col = int(info["column"])
            val = entry.get()
            if val == "" or val == "0":
                val = 0
            else:
                val = 1
            if val == 1:
                ones.append((row, col))
        return ones


    # Checks names of existing codes, creates unique name if needed    
    def get_suitable_name(self, name, existing_names):
        if name not in existing_names:
            return name

        c = 1
        while True:
            if name + "(" + str(c) + ")" not in existing_names:
                return name + "(" + str(c) + ")"
            else:
                c += 1

    # Done button clicked, save the new code
    def done_btn_pressed(self):
        global existing_codes, codes_file_name, selected_code
        name = self.name_entry.get()
        width = self.width_str.get()
        height = self.height_str.get()

        valid, width, height  = self.validate_inputs(width, height)
        if not valid:
            print("Invalid width/height")
            self.master.destroy()
            return
        matrix = self.get_matrix(self.matrix_frame)

        server_values = {}
        for server in range(width):
            server_values[server] = [0]*height
            
        for i in range(height):
            for j in range(width):
                if (i, j) in matrix:
                    server_values[j][i] = 1
                else:
                    server_values[j][i] = 0

        if name == "":
            name = "Unnamed"

        name = self.get_suitable_name(name, existing_codes.keys())

        dict_item = {"name": name, "width": width, "height": height, "matrix": matrix, "server_values": server_values}
        existing_codes[name] = dict_item
        selected_code = dict_item
        
        codes_file = open(codes_file_name, "a")
        codes_file.write(str(dict_item) + "\n")
        codes_file.close()

        print(dict_item)
        self.master.destroy()

        return list(existing_codes.keys())

    # Checks if width & height inputs are suitable
    def validate_inputs(self, width, height):
        try:
            width_int = int(width)
            height_int = int(height)
            if width_int > 0 and height_int > 0:
                return True, width_int, height_int
            else:
                print("Negative size is no good")
        except ValueError:
            return False, None, None


    # Called when user inputs height or width. Then checks if input is valid and creates an empty matrix if valid 
    def size_changed(self, index, value, op):
        width = self.width_str.get()
        height = self.height_str.get()
        
        for entry in self.matrix_frame.winfo_children():
            entry.destroy()

        valid, width, height  = self.validate_inputs(width, height)
        if not valid:
            print("not valid")
            return
        
        for i in range(width):
            for j in range(height):
                b = Entry(self.matrix_frame, text="", width=3)
                b.grid(row=j, column=i)
        self.matrix_frame.pack()

# Frame for the graph        
class GraphFrame(Frame):
    def __init__(self, parent, controller):
        Frame.__init__(self, parent)
        
        canvas = FigureCanvasTkAgg(fig, self)
        canvas.show()
        canvas.get_tk_widget().pack(side=BOTTOM, fill=BOTH, expand=True)

        toolbar = NavigationToolbar2TkAgg(canvas, self)
        toolbar.update()
        canvas._tkcanvas.pack(side=TOP, fill=BOTH, expand=True)

# Code settings window                
class SettingsWindow:
    def __init__(self, master, code):
        self.code = code
        self.master = master
        self.name_frame = Frame(self.master)
        self.button_frame = Frame(self.master)
        self.batch_boolean = code.run_as_batch

        # Code matrix & routing table
        self.info_frame = Frame(self.master)
        self.matrix_frame = Frame(self.info_frame, highlightbackground="gray50", highlightcolor="gray50", highlightthickness=2, bd= 4)
        self.matrix_frame.grid(column=1, row=1)
        self.routing_canvas = Canvas(self.info_frame, width=50, height=100)
        self.routing_canvas.grid(column=2, row=1)
        self.routing_frame = Frame(self.routing_canvas)
        self.settings_frame = Frame(self.master)        
        
        self.draw_settings()

    # Used to draw/redraw settings frame
    def draw_settings(self):
        # Get selected code info
        width = self.code.server_amount
        height = self.code.file_amount
        name = self.code.name
        
        # Clear the frame
        for entry in self.settings_frame.winfo_children():
            entry.destroy()

        # Settings frame & content
        self.settings_label = Label(self.settings_frame, text="Settings for: " + self.code.name)
        self.settings_label.grid(column=0, row=0)
        self.batch_label = Label(self.settings_frame, text="Run code as: ")
        self.batch_label.grid(column=0, row=1)
        self.batch = IntVar()
        if self.code.run_as_batch:
            self.batch.set(1)
        self.async_radio = Radiobutton(self.settings_frame, text="Async", variable=self.batch, value=0, command=partial(self.batch_changed, 0, self.code))
        self.async_radio.grid(column=1, row=1)
        self.batch_radio = Radiobutton(self.settings_frame, text="Batch", variable=self.batch, value=1, command=partial(self.batch_changed, 1, self.code))
        self.batch_radio.grid(column=2, row=1)

        # Settings based on self.code.batch (async if batch == 0, regular batch if 1)
        if self.batch.get() == 1:
            if self.code.batch_table == {}:
                # No table, allow generation
                self.no_table_label = Label(self.settings_frame, text="No batch table for selected code.\nHit Generate to create the table.")
                self.no_table_label.grid(column=0, row=2)
                self.generate_button = Button(self.settings_frame, text="Generate", command=partial(self.generate_batch_button_clicked, self.code))
                self.generate_button.grid(column=1, row=2)
            else:
                # Tables generated, display batch_size selection
                sizes = list(self.code.batch_table.keys())
                print(sizes)
                print(self.code.batch_size)
                selected_size = self.code.batch_size
                self.batch_size_label = Label(self.settings_frame, text="Batch size: ")
                self.batch_size_label.grid(column=0, row=2)
                self.batch_dropdown_str = StringVar()
                self.batch_dropdown = Combobox(self.settings_frame, textvariable=self.batch_dropdown_str, width=5)
                self.batch_dropdown['values'] = sizes
                self.batch_dropdown.current(sizes.index(selected_size))
                self.batch_dropdown_str.trace('w', self.batch_dropdown_item_selected)
                self.batch_dropdown.grid(column=1, row=2)
        # Async settings
        else:
            # Skip distance
            self.skip_distance_label = Label(self.settings_frame, text="Skip distance: ")
            self.skip_distance_label.grid(column=0, row=2)
            self.skip_distance_str = StringVar()
            self.skip_distance_str.set(str(self.code.skip_distance))
            self.skip_distance_str.trace('w', self.skip_distance_changed)
            self.skip_distance_entry = Entry(self.settings_frame, textvariable=self.skip_distance_str)
            self.skip_distance_entry.grid(column=1, row=2)

        # TTL choices   
        self.lifetime_mode_label = Label(self.settings_frame, text="Request lifetime distribution: ")
        self.lifetime_mode_label.grid(column=0, row=3)
        self.lifetime_mode = IntVar()
        if self.code.TTL_uniform:
            self.lifetime_mode.set(0)
        elif self.code.TTL_poisson:
            self.lifetime_mode.set(1)
        elif self.code.TTL_exponential:
            self.lifetime_mode.set(2)
        elif self.code.TTL_gaussian:
            self.lifetime_mode.set(3)
            
        self.uniform_radio = Radiobutton(self.settings_frame, text="Uniform", variable=self.lifetime_mode, value=0, command=partial(self.TTL_mode_changed, 0, self.code))
        self.uniform_radio.grid(column=1, row=3)
        self.poisson_radio = Radiobutton(self.settings_frame, text="Poisson", variable=self.lifetime_mode, value=1, command=partial(self.TTL_mode_changed, 1, self.code))
        self.poisson_radio.grid(column=2, row=3)
        self.exp_radio = Radiobutton(self.settings_frame, text="Exponential", variable=self.lifetime_mode, value=2, command=partial(self.TTL_mode_changed, 2, self.code))
        self.exp_radio.grid(column=3, row=3)
        self.exp_radio = Radiobutton(self.settings_frame, text="Gaussian", variable=self.lifetime_mode, value=3, command=partial(self.TTL_mode_changed, 3, self.code))
        self.exp_radio.grid(column=4, row=3)
        

        # TTL choice based settings
        if self.code.TTL_uniform:
            self.TTL_label = Label(self.settings_frame, text="Request lifetime: ")
            self.TTL_label.grid(column=0, row=4)
            self.min_TTL_str = StringVar()
            self.min_TTL_str.set(str(self.code.min_TTL))
            self.min_TTL_str.trace('w', self.min_TTL_changed)
            self.min_TTL_entry = Entry(self.settings_frame, textvariable=self.min_TTL_str)
            self.min_TTL_entry.grid(column=1, row=4)
            self.TTL_filler = Label(self.settings_frame, text=" to ")
            self.TTL_filler.grid(column=2, row=4)
            self.max_TTL_str = StringVar()
            self.max_TTL_str.set(str(self.code.max_TTL))
            self.max_TTL_str.trace('w', self.max_TTL_changed)
            self.max_TTL_entry = Entry(self.settings_frame, textvariable=self.max_TTL_str)
            self.max_TTL_entry.grid(column=3, row=4)
            self.TTL_ender = Label(self.settings_frame, text=" seconds")
            self.TTL_ender.grid(column=4, row=4)
        elif self.code.TTL_poisson:
            self.TTL_lambda_label = Label(self.settings_frame, text="Request lifetime lambda: ")
            self.TTL_lambda_label.grid(column=0, row=4)
            self.TTL_lambda_str = StringVar()
            self.TTL_lambda_str.set(str(self.code.TTL_lambda))
            self.TTL_lambda_str.trace('w', self.TTL_lambda_changed)
            self.TTL_lambda_entry = Entry(self.settings_frame, textvariable=self.TTL_lambda_str)
            self.TTL_lambda_entry.grid(column=1, row=4)
        elif self.code.TTL_exponential:
            self.TTL_scale_label = Label(self.settings_frame, text="Request lifetime scale: ")
            self.TTL_scale_label.grid(column=0, row=4)
            self.TTL_scale_str = StringVar()
            self.TTL_scale_str.set(str(self.code.TTL_scale))
            self.TTL_scale_str.trace('w', self.TTL_scale_changed)
            self.TTL_scale_entry = Entry(self.settings_frame, textvariable=self.TTL_scale_str)
            self.TTL_scale_entry.grid(column=1, row=4)
        elif self.code.TTL_gaussian:
            self.TTL_mu_label = Label(self.settings_frame, text="Request lifetime mu: ")
            self.TTL_mu_label.grid(column=0, row=4)
            self.TTL_mu_str = StringVar()
            self.TTL_mu_str.set(str(self.code.TTL_mu))
            self.TTL_mu_str.trace('w', self.TTL_mu_changed)
            self.TTL_mu_entry = Entry(self.settings_frame, textvariable=self.TTL_mu_str)
            self.TTL_mu_entry.grid(column=1, row=4)
            self.TTL_filler = Label(self.settings_frame, text=", sigma: ")
            self.TTL_filler.grid(column=2, row=4)
            self.TTL_sigma_str = StringVar()
            self.TTL_sigma_str.set(str(self.code.TTL_sigma))
            self.TTL_sigma_str.trace('w', self.TTL_sigma_changed)
            self.TTL_sigma_entry = Entry(self.settings_frame, textvariable=self.TTL_sigma_str)
            self.TTL_sigma_entry.grid(column=3, row=4)

        # Server capacity
        self.server_capacity_label = Label(self.settings_frame, text="Server capacity: ")
        self.server_capacity_label.grid(column=0, row=5)
        self.server_capacity_str = StringVar()
        self.server_capacity_str.set(str(self.code.server_capacity))
        self.server_capacity_str.trace('w', self.server_capacity_changed)
        self.server_capacity_entry = Entry(self.settings_frame, textvariable=self.server_capacity_str)
        self.server_capacity_entry.grid(column=1, row=5)

        # Gap lambdas
        self.gap_lambdas_frame = Frame(self.settings_frame)
        self.dist_label = Label(self.settings_frame, text="Request arrival gap scales: ")
        self.dist_label.grid(column=0, row=6)
        self.individual_poisson_entries = []
        for file in range(height):
            file_name = Label(self.gap_lambdas_frame, text="F"+str(file+1), width=4, bg="gray80")
            file_name.grid(row=0, column=file)
            file_value_str = StringVar()
            if file in self.code.request_gap_lambdas.keys():
                file_value_str.set(str(self.code.request_gap_lambdas[file]))
            else:
                file_value_str.set("")
                
            file_value_str.trace('w', lambda name, index, mode, file=file : self.file_value_changed(file))
            file_value = Entry(self.gap_lambdas_frame, textvariable=file_value_str, width=5)
            file_value.grid(row=1, column=file)
                
            self.individual_poisson_entries.append(file_value_str)
        self.gap_lambdas_frame.grid(column=1, row=6)
                
        # Set width & height values
        self.width_label = Label(self.info_frame, text=str(width) + " Servers")
        self.width_label.grid(column=1, row=0)
        self.height_label = Label(self.info_frame, text=str(height) + " Files")
        self.height_label.grid(column=0, row=1)

        # Fill matrix and row & column names
        for j in range(height):
            row_name = Label(self.matrix_frame, text="File "+str(j+1), width=3, bg="gray80")
            row_name.grid(row=j+1, column=0)
            
        for i in range(width):
            col_name = Label(self.matrix_frame, text="S"+str(i+1), width=3, bg="gray80")
            col_name.grid(row=0, column=i+1)
            
            for j in range(height):
                val = "0"
                if (j, i) in selected_code["matrix"]:
                    val = "1"
                    
                b = Label(self.matrix_frame, text=val, width=3)
                b.grid(row=j+1, column=i+1)

        # Check if routing table exists for selected code
        if name not in list(existing_routing_tables.keys()):
            # No table, allow generation
            no_table_label = Label(self.routing_frame, text="No routing table for selected code.\nHit Generate to create the table.")
            no_table_label.grid(column=0, row=0)
            generate_button = Button(self.routing_frame, text="Generate", command=partial(self.generate_button_clicked, self.code))
            generate_button.grid(column=0, row=1)
        else:
            # Table exists, display it
            files = list(existing_routing_tables[name]["table"].keys())
            b = Label(self.routing_frame, text="Routing Table")
            b.grid(row=0, column=0)
            for file in files:
                # Make routing table look pretty and display it
                server_combinations = existing_routing_tables[name]["table"][file]
                pretty_combinations = []
                for combination in server_combinations:
                    combination_str = ""
                    for server in combination:
                        combination_str += "+S" + str(server + 1)
                    pretty_combinations.append(combination_str[1:])
                pretty_str = ""
                for combination in pretty_combinations:
                    pretty_str += "[" + combination + "], "
                
                b = Label(self.routing_frame, text="File " + str(file + 1) + ": " + pretty_str[:-2])
                b.grid(row=file + 1, column=0)

        self.routing_frame.pack()
        #self.routing_canvas.config(scrollregion=self.routing_canvas.bbox("all"))
        self.info_frame.pack()
        self.settings_frame.pack(anchor=W)


    # Generates routing table for selected code when Generate button is pressed
    def generate_button_clicked(self, code):
        code.generate_routing_table()
        self.draw_settings()

    # Generates batch table for the code
    def generate_batch_button_clicked(self, code):
        code.generate_batch_table()
        self.draw_settings()

    # Changes settings based on selected TTL distribution
    def TTL_mode_changed(self, value, code):
        if value == 0:
            self.code.set_TTL_uniform()
        elif value == 1:
            self.code.set_TTL_poisson()
        elif value == 2:
            self.code.set_TTL_exponential()
        else:
            self.code.set_TTL_gaussian()
        self.draw_settings()
        
    # Changes settings based on selected batch type
    def batch_changed(self, value, code):
        if value == 0:
            self.code.run_as_batch = False
        else:
            self.code.run_as_batch = True
        self.draw_settings()

    # Changes settings based on selected distribution type
    def dist_changed(self, value, code):
        if value == 0:
            self.code.uniform = False
        else:
            self.code.uniform = True
        self.draw_settings()

    # Changes code request_probabilities
    def file_value_changed(self, file):
        entry = self.individual_poisson_entries[file]
        try:
            value = entry.get()
            if value[-1] == ".":
                return
            self.code.request_gap_lambdas[file] = float(value)
            self.draw_settings()
        except ValueError:
            print("Invalid poisson input")
        
    # Batch size selected
    def batch_dropdown_item_selected(self, index, value, op):
        value = self.batch_dropdown_str.get()
        self.code.batch_size = int(value)
        self.draw_settings()

    # Skip distance changed
    def skip_distance_changed(self, index, value, op):
        try:
            value = int(self.skip_distance_str.get())
            self.code.skip_distance = value
            print("Set skip distance as: " + str(value))
            self.draw_settings()
        except ValueError:
            print("Invalid skip distance")
        
    # TTL lambda changed
    def TTL_lambda_changed(self, index, value, op):
        try:
            value = self.TTL_lambda_str.get()
            if value[-1] == ".":
                return
            value = float(value)
            self.code.TTL_lambda = value
            print("Set TTL_lambda as: " + str(value))
            self.draw_settings()
        except ValueError:
            print("Invalid TTL_lambda")

    # TTL mu changed
    def TTL_mu_changed(self, index, value, op):
        try:
            value = self.TTL_mu_str.get()
            if value[-1] == ".":
                return
            value = float(value)
            self.code.TTL_mu = value
            print("Set TTL_mu as: " + str(value))
            self.draw_settings()
        except ValueError:
            print("Invalid TTL_mu")

    # TTL mu changed
    def TTL_sigma_changed(self, index, value, op):
        try:
            value = self.TTL_sigma_str.get()
            if value[-1] == ".":
                return
            value = float(value)
            self.code.TTL_sigma = value
            print("Set TTL_sigma as: " + str(value))
            self.draw_settings()
        except ValueError:
            print("Invalid TTL_sigma")

    # TTL scale changed
    def TTL_scale_changed(self, index, value, op):
        try:
            value = self.TTL_scale_str.get()
            if value[-1] == ".":
                return
            value = float(value)
            self.code.TTL_scale = value
            print("Set TTL_scale as: " + str(value))
            self.draw_settings()
        except ValueError:
            print("Invalid TTL_scale")
            
    # min TTL changed
    def min_TTL_changed(self, index, value, op):
        try:
            value = self.min_TTL_str.get()
            if value[-1] == ".":
                return
            value = float(value)
            self.code.min_TTL = value
            print("Set min_TTL distance as: " + str(value))
            self.draw_settings()
        except ValueError:
            print("Invalid min_TTL")

    # max_TTL changed
    def max_TTL_changed(self, index, value, op):
        try:
            value = self.max_TTL_str.get()
            if value[-1] == ".":
                return
            value = float(value)
            self.code.max_TTL = value
            print("Set max_TTL as: " + str(value))
            self.draw_settings()
        except ValueError:
            print("Invalid max_TTL")

    # Server capacity changed
    def server_capacity_changed(self, index, value, op):
        try:
            value = int(self.server_capacity_str.get())
            self.code.server_capacity = value
            self.code.update_server_max_capacity()
            print("Set server capacity as: " + str(value))
            self.draw_settings()
        except ValueError:
            print("Invalid server capacity")

    # New request lambda changed
    def request_poisson_lambda_changed(self, index, value, op):
        try:
            value = self.request_poisson_lambda_str.get()
            if value[-1] == ".":
                return
            value = float(value)
            self.code.request_poisson_lambda = value
            self.code.skip_distance = value * 2
            print("Set request_poisson_lambda as: " + str(value))
            self.draw_settings()
        except ValueError:
            print("Invalid request_poisson_lambda")
            
# Current requests window     
class CurrentRequests:
    def __init__(self, master, code):
        self.code = code
        self.master = master
        self.request_frame = Frame(self.master)

        # Code name
        name_str = self.code.name
        if self.code.run_as_batch:
            name_str += " batch, size: " + str(self.code.batch_size)
        else:
            name_str += " async, skip: " + str(self.code.skip_distance)

        # Frame contents
        self.code_name = Label(self.request_frame, text="Code name: ")
        self.code_name.grid(row=0, column=0)
        self.code_name_str = Label(self.request_frame, text=name_str)
        self.code_name_str.grid(row=0, column=1)

        self.file = Label(self.request_frame, text="File", width=4, bg="gray80")
        self.file.grid(row=1, column=0)
        self.TTL = Label(self.request_frame, text="TTL", width=4, bg="gray80")
        self.TTL.grid(row=1, column=1)
        self.servers = Label(self.request_frame, text="Servers", width=8, bg="gray80")
        self.servers.grid(row=1, column=2)

        self.rows = []

        for i in range(self.code.batch_size):
            file = Label(self.request_frame, text="")
            file.grid(row = i+2, column = 0)
            TTL = Label(self.request_frame, text="")
            TTL.grid(row = i+2, column = 1)
            servers = Label(self.request_frame, text="")
            servers.grid(row = i+2, column = 2)
            self.rows.append([file, TTL, servers])

        self.request_frame.pack()
        self.draw_info()

    # Used to draw/redraw info frame
    def draw_info(self):
        self.master.after(100, self.draw_info)
        requests = self.get_current_requests()
        
        #Clear the current_requests frame
        for row in self.rows:
            for label in row:
                label["text"] = ""
        
        for i in range(len(requests)):
            request = requests[i]
            row = self.rows[i]
            row[0]["text"] = request["file"]
            row[1]["text"] = request["TTL"]
            row[2]["text"] = str(request["servers"])

    # Gets current requests info from code.servers
    def get_current_requests(self):
        requests = []
        for request in self.code.current_requests:
            requests.append({"file": "F"+str(request["requested_file"]+1), "TTL": str(int(request["TTL"])), "servers": request["servers"]})
        return requests

        
# Code info window                
class InfoWindow:
    def __init__(self, master, code):
        self.code = code
        self.master = master
        self.info_frame = Frame(self.master)
        
        # Code name
        name_str = code.name
        if code.run_as_batch:
            name_str += " batch, size: " + str(code.batch_size)
        else:
            name_str += " async, skip: " + str(code.skip_distance)

        # Frame contents
        self.code_name = Label(self.info_frame, text="Code name: ")
        self.code_name.grid(row=0, column=0)
        self.code_name_str = Label(self.info_frame, text=name_str)
        self.code_name_str.grid(row=0, column=1)

        self.request_queue_size_label = Label(self.info_frame, text="Queue size: ")
        self.request_queue_size_label.grid(row=1, column=0)
        self.request_queue_size_str = StringVar()
        self.request_queue_size = Label(self.info_frame, textvariable=self.request_queue_size_str)
        self.request_queue_size.grid(row=1, column=1)

        self.current_requests_label = Label(self.info_frame, text="Current requests: ")
        self.current_requests_label.grid(row=2, column=0)
        self.current_requests_str = StringVar()
        self.current_requests = Label(self.info_frame, textvariable=self.current_requests_str)
        self.current_requests.grid(row=2, column=1)

        self.completed_requests_label = Label(self.info_frame, text="Completed requests: ")
        self.completed_requests_label.grid(row=3, column=0)
        self.completed_requests_str = StringVar()
        self.completed_requests = Label(self.info_frame, textvariable=self.completed_requests_str)
        self.completed_requests.grid(row=3, column=1)

        self.requests_per_second_label = Label(self.info_frame, text="Requests/sec: ")
        self.requests_per_second_label.grid(row=4, column=0)
        self.requests_per_second_str = StringVar()
        self.requests_per_second = Label(self.info_frame, textvariable=self.requests_per_second_str)
        self.requests_per_second.grid(row=4, column=1)

        self.servers_in_use_label = Label(self.info_frame, text="Servers in use: ")
        self.servers_in_use_label.grid(row=5, column=0)
        self.servers_in_use_str = StringVar()
        self.servers_in_use = Label(self.info_frame, textvariable=self.servers_in_use_str)
        self.servers_in_use.grid(row=5, column=1)

        self.servers_at_capacity_label = Label(self.info_frame, text="Servers at capacity: ")
        self.servers_at_capacity_label.grid(row=6, column=0)
        self.servers_at_capacity_str = StringVar()
        self.servers_at_capacity = Label(self.info_frame, textvariable=self.servers_at_capacity_str)
        self.servers_at_capacity.grid(row=6, column=1)

        self.servers_capacity_percent_label = Label(self.info_frame, text="Servers capacity %: ")
        self.servers_capacity_percent_label.grid(row=7, column=0)
        self.servers_capacity_percent_str = StringVar()
        self.servers_capacity_percent = Label(self.info_frame, textvariable=self.servers_capacity_percent_str)
        self.servers_capacity_percent.grid(row=7, column=1)

        self.queue_time_mean_label = Label(self.info_frame, text="Mean queue time: ")
        self.queue_time_mean_label.grid(row=8, column=0)
        self.queue_time_mean_str = StringVar()
        self.queue_time_mean = Label(self.info_frame, textvariable=self.queue_time_mean_str)
        self.queue_time_mean.grid(row=8, column=1)

        self.queue_time_median_label = Label(self.info_frame, text="Median queue time: ")
        self.queue_time_median_label.grid(row=9, column=0)
        self.queue_time_median_str = StringVar()
        self.queue_time_median = Label(self.info_frame, textvariable=self.queue_time_median_str)
        self.queue_time_median.grid(row=9, column=1)

        self.queue_time_stdev_label = Label(self.info_frame, text="Queue time stdev: ")
        self.queue_time_stdev_label.grid(row=10, column=0)
        self.queue_time_stdev_str = StringVar()
        self.queue_time_stdev = Label(self.info_frame, textvariable=self.queue_time_stdev_str)
        self.queue_time_stdev.grid(row=10, column=1)

        self.queue_time_variance_label = Label(self.info_frame, text="Queue time variance: ")
        self.queue_time_variance_label.grid(row=11, column=0)
        self.queue_time_variance_str = StringVar()
        self.queue_time_variance = Label(self.info_frame, textvariable=self.queue_time_variance_str)
        self.queue_time_variance.grid(row=11, column=1)

        self.queue_time_max_label = Label(self.info_frame, text="Max queue time: ")
        self.queue_time_max_label.grid(row=12, column=0)
        self.queue_time_max_str = StringVar()
        self.queue_time_max = Label(self.info_frame, textvariable=self.queue_time_max_str)
        self.queue_time_max.grid(row=12, column=1)

        self.running_time_label = Label(self.info_frame, text="Code running time: ")
        self.running_time_label.grid(row=13, column=0)
        self.running_time_str = StringVar()
        self.running_time = Label(self.info_frame, textvariable=self.running_time_str)
        self.running_time.grid(row=13, column=1)


        # TODO: rows
        self.buffer_label1 = Label(self.info_frame, text=" ")
        self.buffer_label1.grid(row = 14, column = 0)
        
        # Server capacity table
        self.servers_label = Label(self.info_frame, text="Server: ")
        self.servers_label.grid(row=15, column=1)
        self.cap_label = Label(self.info_frame, text="Capacity %: ")
        self.cap_label.grid(row=16, column=1)
        self.cap_values = []
        for i in range(self.code.server_amount):
            col_name = Label(self.info_frame, text="S"+str(i+1), width=4, bg="gray80")
            col_name.grid(row=15, column=i+2)
            col_value_str = StringVar()
            col_value = Label(self.info_frame, textvariable=col_value_str)
            col_value.grid(row=16, column=i+2)
            self.cap_values.append(col_value_str)

        self.buffer_label2 = Label(self.info_frame, text=" ")
        self.buffer_label2.grid(row = 17, column = 0)
        
        # Request queue file amounts
        self.files_label = Label(self.info_frame, text="File: ")
        self.files_label.grid(row=18, column=1)
        self.amount_label = Label(self.info_frame, text="Requests in queue: ")
        self.amount_label.grid(row=19, column=1)
        self.amount_values = []
        for i in range(self.code.file_amount):
            col_name = Label(self.info_frame, text="F"+str(i+1), width=4, bg="gray80")
            col_name.grid(row=18, column=i+2)
            col_value_str = StringVar()
            col_value = Label(self.info_frame, textvariable=col_value_str)
            col_value.grid(row=19, column=i+2)
            self.amount_values.append(col_value_str)

        self.buffer_label3 = Label(self.info_frame, text=" ")
        self.buffer_label3.grid(row = 20, column = 0)
        
        # Request queue file amounts
        self.files_label2 = Label(self.info_frame, text="File: ")
        self.files_label2.grid(row=21, column=1)
        self.amount_label2 = Label(self.info_frame, text="Active requests: ")
        self.amount_label2.grid(row=22, column=1)
        self.amount_values2 = []
        for i in range(self.code.file_amount):
            col_name = Label(self.info_frame, text="F"+str(i+1), width=4, bg="gray80")
            col_name.grid(row=21, column=i+2)
            col_value_str = StringVar()
            col_value = Label(self.info_frame, textvariable=col_value_str)
            col_value.grid(row=22, column=i+2)
            self.amount_values2.append(col_value_str)
        
        self.info_frame.pack()
        self.draw_info()

    # Used to draw/redraw info frame
    def draw_info(self):
        self.master.after(100, self.draw_info)
            
        self.request_queue_size_str.set(str(self.code.statistics["request_queue_size"][-1]))
        self.current_requests_str.set(str(self.code.statistics["current_requests"][-1]))
        self.completed_requests_str.set(str(self.code.statistics["completed_requests"][-1]))
        self.requests_per_second_str.set(str(round(self.code.statistics["requests_per_second"][-1], 2)))
        self.servers_in_use_str.set(str(self.code.statistics["servers_in_use"][-1]))
        self.servers_at_capacity_str.set(str(self.code.statistics["servers_at_capacity"][-1]))
        self.servers_capacity_percent_str.set(str(round(self.code.statistics["servers_capacity_percent"][-1], 2)))
        self.queue_time_mean_str.set(str(round(self.code.statistics["queue_time_mean"][-1], 2)))
        self.queue_time_median_str.set(str(round(self.code.statistics["queue_time_median"][-1], 2)))
        self.queue_time_stdev_str.set(str(round(self.code.statistics["queue_time_stdev"][-1], 2)))
        self.queue_time_variance_str.set(str(round(self.code.statistics["queue_time_variance"][-1], 2)))
        self.queue_time_max_str.set(str(self.code.statistics["queue_time_max"][-1]))
        self.running_time_str.set(str(self.code.running_time / 100))
        
        server_capacities = self.get_server_capacities()
        for server in server_capacities.keys():
            self.cap_values[server].set(server_capacities[server])

        request_queue_info = self.get_request_queue_info()
        for file in request_queue_info.keys():
            self.amount_values[file].set(request_queue_info[file])

        current_requests_info = self.get_current_requests_info()
        for file in current_requests_info.keys():
            self.amount_values2[file].set(current_requests_info[file])


        self.info_frame.pack()

    # Sums the amounts of requests in code.request_queue for each file
    def get_request_queue_info(self):
        file_count = {x: 0 for x in range(self.code.file_amount)}
        for request in self.code.request_queue:
            file_count[request["requested_file"]] += 1
        return file_count

    # Gets server capacity per server from code.servers
    def get_server_capacities(self):
        server_capacities = {}
        for server in self.code.servers.keys():
            server_capacities[server] = round((self.code.servers[server]["requests"] / self.code.server_capacity)*100, 1)
        return server_capacities

    # Gets current requests info from code.servers
    def get_current_requests_info(self):
        file_count = {x: 0 for x in range(self.code.file_amount)}
        for request in self.code.current_requests:
            file_count[request["requested_file"]] += 1
        return file_count
    

main()
